from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, Field, model_validator


class CaptureFileConfig(BaseModel):
    hash_algorithm: Literal["md5", "sha1", "sha256", "sha3"] | None = None
    copy_: bool = Field(default=False, alias="copy")
    move: bool = False

    @model_validator(mode="after")  # type: ignore
    def exclusive_copy_move(cls, model: CaptureFileConfig) -> CaptureFileConfig:  # noqa: N805
        if model.copy_ and model.move:
            msg = "Only one of copy or move can be true"
            raise ValueError(msg)
        return model



from __future__ import annotations

import logging
import subprocess
import sys

if sys.version_info < (3, 11):
    from datetime import timezone as _timezone

    UTC = _timezone.utc

else:
    from datetime import UTC
from datetime import datetime
from pathlib import Path

from pydantic import BaseModel, ConfigDict, Field

from capsula.context import Context
from capsula.file import CaptureFileConfig  # noqa: TCH001 for pydantic
from capsula.globalvars import set_capsule_dir

logger = logging.getLogger(__name__)


def to_hyphen_case(string: str) -> str:
    return string.replace("_", "-")


class GitConfig(BaseModel):
    repositories: dict[str, Path] = Field(default_factory=dict)


class CaptureConfig(BaseModel):
    """Configuration for the capture command."""

    model_config = ConfigDict(
        alias_generator=to_hyphen_case,
        populate_by_name=True,
        extra="forbid",
    )

    vault_directory: Path
    capsule_template: str

    # Whether to include the Capsula version in the output file.
    # include_capsula_version: bool = True

    include_cpu: bool = True

    pre_capture_commands: list[str] = Field(default_factory=list)

    environment_variables: list[str] = Field(default_factory=list)

    files: dict[Path, CaptureFileConfig] = Field(default_factory=dict)

    git: GitConfig = Field(default_factory=GitConfig)

    _capsule_directory: Path | None = None

    @property
    def capsule(self) -> Path:
        if self._capsule_directory is None:
            self._capsule_directory = self.vault_directory / datetime.now(UTC).astimezone().strftime(
                self.capsule_template,
            )
        return self._capsule_directory


def capture(*, config: CaptureConfig) -> Context:
    """Capture the context."""
    logger.debug(f"Capture config: {config}")

    logger.info(f"CWD: {Path.cwd()}")
    for command in config.pre_capture_commands:
        logger.info(f"Running pre-capture command: {command!r}")
        try:
            result = subprocess.run(command, shell=True, text=True, capture_output=True, check=True)  # noqa: S602
        except subprocess.CalledProcessError:  # noqa: PERF203
            logger.exception(f"Pre-capture command failed: {command}")
            raise
        else:
            logger.info(f"Pre-capture command result: {result}")

    logger.info("Capturing the context.")

    try:
        config.capsule.mkdir(parents=True, exist_ok=False)
    except FileExistsError:
        logger.exception(f"Capsule already exists: {config.capsule}")
        raise
    else:
        set_capsule_dir(config.capsule)

    ctx = Context.capture(config)

    # Write the context to the output file.
    with (config.capsule / "context.json").open("w") as output_file:
        output_file.write(ctx.model_dump_json(indent=4))

    return ctx



import os
from pathlib import Path


def set_capsule_dir(path: Path | str) -> None:
    """Set the capsule directory used by the monitor CLI."""
    os.environ["CAPSULE_DIR"] = str(path)


def get_capsule_dir() -> Path:
    """Get the capsule directory used by the monitor CLI."""
    try:
        return Path(os.environ["CAPSULE_DIR"])
    except KeyError as e:
        msg = "CAPSULE_DIR is not set. Run `capsula monitor <args>` to set it."
        raise RuntimeError(msg) from e



from __future__ import annotations

import os
import platform as pf
import sys
from abc import ABC, abstractmethod
from pathlib import Path
from shutil import copyfile, move
from typing import TYPE_CHECKING, Any, Literal

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

from cpuinfo import get_cpu_info
from git.repo import Repo
from pydantic import BaseModel, Field

from capsula.hash import compute_hash

if TYPE_CHECKING:
    from capsula.capture import CaptureConfig


class ContextItem(BaseModel, ABC):
    """Base class for context items."""

    @classmethod
    @abstractmethod
    def capture(cls, config: CaptureConfig) -> Self | dict[Any, Self]:
        """Capture the context item."""
        raise NotImplementedError


class Architecture(ContextItem):
    bits: str
    linkage: str

    @classmethod
    def capture(cls, _: CaptureConfig) -> Self:
        return cls(
            bits=pf.architecture()[0],
            linkage=pf.architecture()[1],
        )


class PythonInfo(ContextItem):
    executable_architecture: Architecture
    build_no: str
    build_date: str
    compiler: str
    branch: str
    implementation: str
    version: str

    @classmethod
    def capture(cls, config: CaptureConfig) -> Self:
        return cls(
            executable_architecture=Architecture.capture(config),
            build_no=pf.python_build()[0],
            build_date=pf.python_build()[1],
            compiler=pf.python_compiler(),
            branch=pf.python_branch(),
            implementation=pf.python_implementation(),
            version=pf.python_version(),
        )


class Platform(ContextItem):
    """Information about the platform."""

    machine: str
    node: str
    platform: str
    release: str
    version: str
    system: str
    processor: str
    python: PythonInfo

    @classmethod
    def capture(cls, config: CaptureConfig) -> Self:
        return cls(
            machine=pf.machine(),
            node=pf.node(),
            platform=pf.platform(),
            release=pf.release(),
            version=pf.version(),
            system=pf.system(),
            processor=pf.processor(),
            python=PythonInfo.capture(config),
        )


class GitRemote(BaseModel):
    name: str
    url: str


class GitInfo(ContextItem):
    path: Path
    sha: str
    branch: str
    remotes: list[GitRemote]

    @classmethod
    def capture(cls, config: CaptureConfig) -> dict[str, Self]:
        git_infos = {}
        for name, path in config.git.repositories.items():
            repo = Repo(path)
            git_infos[name] = cls(
                path=path,
                sha=repo.head.object.hexsha,
                remotes=[GitRemote(name=remote.name, url=remote.url) for remote in repo.remotes],
                branch=repo.active_branch.name,
            )

            diff = repo.git.diff()
            with (config.capsule / f"{name}.diff").open("w") as diff_file:
                diff_file.write(diff)

        return git_infos


class FileContext(ContextItem):
    hash_algorithm: Literal["md5", "sha1", "sha256", "sha3"] | None
    file_hash: str | None = Field(..., alias="hash")

    @classmethod
    def capture(cls, config: CaptureConfig) -> dict[Path, Self]:
        files = {}
        for path, file_config in config.files.items():
            files[path] = cls(
                hash=compute_hash(path, file_config.hash_algorithm) if file_config.hash_algorithm else None,
                hash_algorithm=file_config.hash_algorithm,
            )
            if file_config.copy_:
                copyfile(path, config.capsule / path.name)
            elif file_config.move:
                move(path, config.capsule / path.name)

        return files


class Context(ContextItem):
    """Execution context to be stored and used later."""

    platform: Platform

    # There are many duplicates between the platform and cpu info.
    # We could remove the duplicates, but it's not worth the effort.
    # We use the default factory to avoid the overhead of getting the CPU info, which is slow.
    cpu: dict | None

    environment_variables: dict[str, str]

    cwd: Path

    git: dict[str, GitInfo]

    files: dict[Path, FileContext]

    @classmethod
    def capture(cls, config: CaptureConfig) -> Self:
        return cls(
            platform=Platform.capture(config),
            cpu=get_cpu_info() if config.include_cpu else None,
            environment_variables={name: os.environ[name] for name in config.environment_variables},
            git=GitInfo.capture(config),
            cwd=Path.cwd(),
            files=FileContext.capture(config),
        )



import hashlib
import logging
from pathlib import Path
from typing import Literal

logger = logging.getLogger(__name__)


def compute_hash(file_path: Path, algorithm: Literal["md5", "sha1", "sha256", "sha3"]) -> str:
    buf_size = 65536  # lets read stuff in 64kb chunks!

    if algorithm == "md5":
        hash_algo = hashlib.md5()  # noqa: S324
    elif algorithm == "sha1":
        hash_algo = hashlib.sha1()  # noqa: S324
    elif algorithm == "sha256":
        hash_algo = hashlib.sha256()
    elif algorithm == "sha3":
        hash_algo = hashlib.sha3_256()
    else:
        msg = "Unknown algorithm"
        raise ValueError(msg)

    with file_path.open("rb") as f:
        while True:
            data = f.read(buf_size)
            if not data:
                break
            hash_algo.update(data)

    return hash_algo.hexdigest()



from __future__ import annotations

import logging
import sys

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib
from pathlib import Path
from typing import TYPE_CHECKING

import click

from capsula._monitor import MonitorConfig, MonitoringHandlerCli
from capsula.capture import CaptureConfig
from capsula.capture import capture as capture_core

if TYPE_CHECKING:
    from collections.abc import Iterable


@click.group()
@click.option(
    "--directory",
    "-C",
    type=click.Path(exists=True, file_okay=False, dir_okay=True, resolve_path=True, path_type=Path),
    default=Path.cwd(),
    help="The working directory to use. Defaults to the current directory.",
)
@click.option(
    "--log-level",
    "-l",
    type=click.Choice(
        [
            "CRITICAL",
            # "FATAL", # FATAL is an alias for CRITICAL
            "ERROR",
            # "WARN", # WARN is an alias for WARNING
            "WARNING",
            "INFO",
            "DEBUG",
            "NOTSET",
        ],
    ),
    default="INFO",
    help="The log level to use. Defaults to INFO.",
)
@click.pass_context
def main(
    ctx: click.Context,
    directory: Path,
    log_level: str,
) -> None:
    ctx.ensure_object(dict)
    logging.basicConfig(level=logging.getLevelNamesMapping()[log_level])

    ctx.obj["directory"] = directory
    click.echo(f"directory: {ctx.obj['directory']}")
    capsula_config_path: Path = ctx.obj["directory"] / "capsula.toml"
    with capsula_config_path.open("rb") as capsula_config_file:
        capsula_config = tomllib.load(capsula_config_file)

    ctx.obj["capsula_config"] = capsula_config


@main.command()
@click.pass_context
def capture(ctx: click.Context) -> None:
    """Capture the context."""
    capsula_capture_config = CaptureConfig(**ctx.obj["capsula_config"]["capture"])
    capture_core(config=capsula_capture_config)


@main.command()
@click.option(
    "--item",
    "-i",
    "items",
    multiple=True,
    default=(),
    help="The item to monitor defined in the capsula.toml file.",
)
@click.argument("args", nargs=-1)
@click.pass_context
def monitor(ctx: click.Context, items: Iterable[str], args: tuple[str]) -> None:
    """Monitor execution."""
    capture_config = CaptureConfig(**ctx.obj["capsula_config"]["capture"])
    capture_core(config=capture_config)

    monitor_config = MonitorConfig(**ctx.obj["capsula_config"]["monitor"])
    handler = MonitoringHandlerCli(capture_config=capture_config, monitor_config=monitor_config)
    pre_run_info = handler.setup(args)
    handler.run_and_teardown(pre_run_info=pre_run_info, items=items)



from __future__ import annotations

import inspect
import logging
import subprocess
import sys
import time

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib
import traceback
import warnings
from abc import ABC, abstractmethod

if sys.version_info < (3, 11):
    from datetime import timezone as _timezone

    UTC = _timezone.utc
else:
    from datetime import UTC
from datetime import datetime, timedelta
from functools import wraps
from pathlib import Path
from shutil import copyfile, move
from typing import TYPE_CHECKING, Any, Generic, Literal, TypeVar

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

from pydantic import BaseModel, Field

from capsula.capture import CaptureConfig
from capsula.capture import capture as capture_core
from capsula.file import CaptureFileConfig  # noqa: TCH001 for pydantic
from capsula.hash import compute_hash

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable, Sequence


logger = logging.getLogger(__name__)


class MonitorItemConfig(BaseModel):
    files: dict[Path, CaptureFileConfig] = Field(default_factory=dict)


class MonitorConfig(BaseModel):
    items: dict[str, MonitorItemConfig] = Field(default_factory=dict, alias="item")


class PreRunInfoBase(BaseModel):
    timestamp: datetime = Field(default_factory=lambda: datetime.now(UTC).astimezone())
    cwd: Path = Field(default_factory=Path.cwd)


class PreRunInfoCli(PreRunInfoBase):
    args: list[str]


class PreRunInfoFunc(PreRunInfoBase):
    source_file: Path | None
    source_line: int | None
    function_name: str
    args: tuple[Any, ...]
    kwargs: dict[str, Any]


class OutputFileInfo(BaseModel):
    hash_algorithm: Literal["md5", "sha1", "sha256", "sha3"] | None
    file_hash: str | None = Field(..., alias="hash")


class PostRunInfoBase(BaseModel):
    timestamp: datetime = Field(default_factory=lambda: datetime.now(UTC).astimezone())
    run_time: timedelta
    files: dict[Path, OutputFileInfo | None] = Field(default_factory=dict)


class PostRunInfoCli(PostRunInfoBase):
    stdout: str
    stderr: str
    exit_code: int


class ExceptionInfo(BaseModel):
    error_type: str
    error_message: str
    error_details: str

    @classmethod
    def from_exception(cls, exc: BaseException) -> ExceptionInfo:
        return cls(
            error_type=type(exc).__name__,
            error_message=str(exc),
            error_details=traceback.format_exc(),
        )


class PostRunInfoFunc(PostRunInfoBase):
    return_value: Any | None = None
    exception_info: ExceptionInfo | None = None


_TPreRunInfo = TypeVar("_TPreRunInfo", bound=PreRunInfoBase)
_TPostRunInfo = TypeVar("_TPostRunInfo", bound=PostRunInfoBase)


class MonitoringHandlerBase(ABC, Generic[_TPreRunInfo, _TPostRunInfo]):
    def __init__(self, *, capture_config: CaptureConfig, monitor_config: MonitorConfig) -> None:
        self.capture_config = capture_config
        self.monitor_config = monitor_config

    @abstractmethod
    def setup_pre_run_info(self, *args: Any, **kwargs: Any) -> _TPreRunInfo:
        ...

    def setup(self, *args: Any, **kwargs: Any) -> _TPreRunInfo:
        """Setup the pre-run info and dump it to a file."""
        pre_run_info = self.setup_pre_run_info(*args, **kwargs)
        self.dump_pre_run_info(pre_run_info)
        return pre_run_info

    def dump_pre_run_info(self, pre_run_info: _TPreRunInfo) -> None:
        with (self.capture_config.capsule / "pre-run-info.json").open("w") as f:
            f.write(pre_run_info.model_dump_json(indent=4))

    @abstractmethod
    def run(
        self,
        pre_run_info: _TPreRunInfo,
        *,
        items: Iterable[str],
        **kwargs: Any,
    ) -> tuple[_TPostRunInfo, BaseException | None]:
        ...

    def run_and_teardown(
        self,
        pre_run_info: _TPreRunInfo,
        *,
        items: Iterable[str],
        **kwargs: Any,
    ) -> tuple[_TPostRunInfo, BaseException | None]:
        post_run_info, exc = self.run(pre_run_info, items=items, **kwargs)
        return self.teardown(post_run_info=post_run_info, items=items), exc

    def teardown(self, post_run_info: _TPostRunInfo, *, items: Iterable[str]) -> _TPostRunInfo:
        post_run_info.files = {}
        for item in items:
            for path, file in self.monitor_config.items[item].files.items():
                logger.debug(f"Capturing file {path}...")
                if not path.exists():
                    logger.warning(f"File {path} does not exist.")
                    post_run_info.files[path] = None
                    continue
                post_run_info.files[path] = OutputFileInfo(
                    hash_algorithm=file.hash_algorithm,
                    hash=compute_hash(path, file.hash_algorithm) if file.hash_algorithm else None,
                )
                if file.copy_:
                    copyfile(path, self.capture_config.capsule / path.name)
                elif file.move:
                    move(path, self.capture_config.capsule / path.name)

        with (self.capture_config.capsule / "post-run-info.json").open("w") as f:
            f.write(post_run_info.model_dump_json(indent=4))

        return post_run_info


class MonitoringHandlerCli(MonitoringHandlerBase[PreRunInfoCli, PostRunInfoCli]):
    def setup_pre_run_info(self, args: Sequence[str]) -> PreRunInfoCli:
        return PreRunInfoCli(args=list(args), cwd=Path.cwd(), timestamp=datetime.now(UTC).astimezone())

    def run(
        self,
        pre_run_info: PreRunInfoCli,
        *,
        items: Iterable[str],  # noqa: ARG002
    ) -> tuple[PostRunInfoCli, None]:
        start_time = time.perf_counter()
        result = subprocess.run(pre_run_info.args, capture_output=True, text=True)  # noqa: S603
        end_time = time.perf_counter()

        return (
            PostRunInfoCli(
                timestamp=datetime.now(UTC).astimezone(),
                run_time=timedelta(seconds=end_time - start_time),
                stdout=result.stdout,
                stderr=result.stderr,
                exit_code=result.returncode,
            ),
            None,
        )


class MonitoringHandlerFunc(MonitoringHandlerBase[PreRunInfoFunc, PostRunInfoFunc]):
    def setup_pre_run_info(
        self,
        func: Callable[..., Any],
        args: Sequence[Any],
        kwargs: dict[str, Any],
    ) -> PreRunInfoFunc:
        try:
            file_path = inspect.getsourcefile(func)
        except TypeError:
            file_path = None
        file_path = Path(file_path) if file_path is not None else None

        try:
            _, first_line_no = inspect.getsourcelines(func)
        except (TypeError, OSError):
            first_line_no = None

        return PreRunInfoFunc(
            cwd=Path.cwd(),
            timestamp=datetime.now(UTC).astimezone(),
            source_file=file_path,
            source_line=first_line_no,
            function_name=func.__name__,
            args=tuple(args),
            kwargs=kwargs,
        )

    def run(
        self,
        pre_run_info: PreRunInfoFunc,
        *,
        items: Iterable[str],  # noqa: ARG002
        func: Callable[..., Any],
    ) -> tuple[PostRunInfoFunc, BaseException | None]:
        start_time = time.perf_counter()
        try:
            ret = func(*pre_run_info.args, **pre_run_info.kwargs)
        except Exception as e:  # noqa: BLE001
            end_time = time.perf_counter()
            exception_info = ExceptionInfo.from_exception(e)
            return (
                PostRunInfoFunc(
                    timestamp=datetime.now(UTC).astimezone(),
                    run_time=timedelta(seconds=end_time - start_time),
                    exception_info=exception_info,
                ),
                e,
            )
        else:
            end_time = time.perf_counter()
            return (
                PostRunInfoFunc(
                    timestamp=datetime.now(UTC).astimezone(),
                    run_time=timedelta(seconds=end_time - start_time),
                    return_value=ret,
                ),
                None,
            )


T = TypeVar("T")
P = ParamSpec("P")


def monitor(
    directory: Path,
    *,
    items: Iterable[str] = (),
    include_return_value: bool = False,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    if isinstance(items, str):
        warnings.warn(
            "Passing a single string to `monitor` is deprecated. Use a tuple instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        items = (items,)

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        capsula_config_path = directory / "capsula.toml"
        with capsula_config_path.open("rb") as capsula_config_file:
            capsula_config = tomllib.load(capsula_config_file)

        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            capture_config = CaptureConfig(**capsula_config["capture"])
            captured_ctx = capture_core(config=capture_config)
            logger.debug(f"Captured context: {captured_ctx}")

            monitor_config = MonitorConfig(**capsula_config["monitor"])
            logger.debug(f"Monitor config: {monitor_config}")

            handler = MonitoringHandlerFunc(capture_config=capture_config, monitor_config=monitor_config)

            pre_run_info = handler.setup(func=func, args=args, kwargs=kwargs)
            post_run_info, exc = handler.run_and_teardown(
                pre_run_info=pre_run_info,
                items=items,
                func=func,
            )
            if exc is not None:
                logger.error(f"Exception occurred: {exc!r}")
                raise exc

            ret = post_run_info.return_value

            if not include_return_value:
                post_run_info.return_value = None

            return ret  # type: ignore

        return wrapper

    return decorator
